@@ -21,12 +21,30 @@ enum blocktype {
 	pointers
 };
 
+/* The kind of label to generate at start or end of a block. */
+enum labeltype {
+	/* Don't generate any label. */
+	LABEL_NONE,
+
+	/* Make a label at the first or last byte in block. */
+	LABEL_FIRST_LAST,
+
+	/* Make a label at the first or one after the last byte in block */
+	LABEL_START_END
+};
+
 struct block {
 	char *name;
 
 	int start;	/* first byte in this block */
 	int end;	/* last byte in this block plus 1 */
 
+	/* The kind of label to generate at the start of this block */
+	enum labeltype start_label;
+
+	/* The kind of label to generate at the end of this block */
+	enum labeltype end_label;
+
 	enum blocktype type;
 
 	struct block *prev;
@@ -37,8 +55,11 @@ extern struct block *blocks;
 
 int block_init(int start, int end);
 
-struct block *block_new(char *name, int start, int end, enum blocktype type);
+struct block *block_new(char *name, int start, int end, enum blocktype type,
+			enum labeltype start_label, enum labeltype end_label);
 
 int block_load_file(char *filename);
 int block_load_line(char *line, char **name, 
-				int *start, int *end, enum blocktype *type);
+				int *start, int *end, enum blocktype *type,
+				enum labeltype *start_label,
+				enum labeltype *end_label);
