@@ -25,6 +25,7 @@
 int a_address = 0;
 int a_labels = 0;
 int a_source = 0;
+int a_undoc = 0;
 int a_org = 0x0100;
 int a_zilog = 0;
 
@@ -76,6 +77,8 @@ void syntax()
 	printf("  -g  --origin=ORG      Specify the starting address of the code (default 100h)\n");
 	printf("  -t  --source          Print the corresponding binary data in hex and ASCII\n");
 	printf("                        for each assembly line in a comment.\n");
+	printf("  -u  --undoc           Insert undocumented instructions directly into the\n");
+	printf("                        disassembly instead of in the comments.\n");
 
 	printf("  -s  --sym-output=FILE Write symbol information into FILE\n");
 	printf("  -S  --sym-input=FILE  Read symbol information from FILE\n");
@@ -107,6 +110,7 @@ void cmdline(int argc, char **argv)
 		{ "origin", required_argument, NULL, 'g'},
 		{ "reduced", no_argument, NULL, 'r' },
 		{ "source", no_argument, NULL, 't' },
+		{ "undoc", no_argument, NULL, 'u' },
 		{ "sym-output", required_argument, NULL, 's' },
 		{ "sym-input", required_argument, NULL, 'S' },
 		{ "block-def", required_argument, NULL, 'b' },
@@ -116,7 +120,7 @@ void cmdline(int argc, char **argv)
 		{ "zilog", no_argument, NULL, 'z' },
 		{ NULL, 0, NULL, 0 }
 	};
-	const char opts[] = "alg:rtS:s:o:hVvzb:";
+	const char opts[] = "alg:rtuS:s:o:hVvzb:";
 	int c,r;
 
 	while((c=getopt_long(argc, argv, opts, longopts, NULL))!=-1) {
@@ -135,6 +139,8 @@ void cmdline(int argc, char **argv)
 				  break;
 			case 't': a_source=1;
 				  break;
+			case 'u': a_undoc=1;
+				  break;
 			case 's': a_symoutput=optarg;
 				  break;
 			case 'S': a_syminput=optarg;
