@@ -17,7 +17,7 @@ along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
-/* Modified by Tomaz Solc <tomaz.solc@tablix.org>, June 2007 */
+/* Modified by Tomaz Solc <tomaz.solc@tablix.org>, June 2017 */
 
 #include <stdio.h>
 #include <string.h>
@@ -176,6 +176,26 @@ int ckrange_rel(int skip, int pc, char *str)
 	}
 }
 
+/* Outputs n bytes as a defb line. Used for invalid/unknown instructions, etc.
+ * If n == -1, then outputs bytes until the end of block.
+ *
+ * Returns number of bytes used. */
+int defb(int n)
+{
+	int i;
+	FP(fx,"defb");
+	for(i=0;t[i]!=-1&&(n==-1||i<n);i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
+	return i;
+}
+
+void before_undoc(int n)
+{
+	if(!a_undoc) {
+		defb(n);
+		FP(fx, "\t;");
+	}
+}
+
 /* Returns number of bytes used */
 int diz80()
 {
@@ -191,8 +211,7 @@ FP(fx,"\t");
 if(t[0] == 0xcb) {
 	a=t[1];
 	if(a==-1) {
-		FP(fx,"defb 0%02xh", t[0]);
-		return(1);
+		return defb(-1);
 	}
 	if(a < 8)				FP(fx,"rlc ");
 	else if((a >= 0x08) && (a < 0x10))	FP(fx,"rrc ");
@@ -201,7 +220,10 @@ if(t[0] == 0xcb) {
 	else if((a >= 0x20) && (a < 0x28))	FP(fx,"sla ");
 	else if((a >= 0x28) && (a < 0x30))	FP(fx,"sra ");
 	/* undocumented instruction SLI (shift left with 1 insertion) */
-	else if((a >= 0x30) && (a < 0x38))	FP(fx,"sli ");
+	else if((a >= 0x30) && (a < 0x38)) {
+		before_undoc(2);
+		FP(fx,"sli ");
+	}
 	else if((a >= 0x38) && (a < 0x40))	FP(fx,"srl ");
 	else if((a >= 0x40) && (a < 0x80))	FP(fx,"bit %d,",(a-0x40)/8);
 	else if((a >= 0x80) && (a < 0xc0))	FP(fx,"res %d,",(a-0x80)/8);
@@ -212,8 +234,7 @@ if(t[0] == 0xcb) {
 } else if(t[0] == 0xed) {
 	a=t[1];
 	if(a==-1) {
-		FP(fx,"defb 0%02xh", t[0]);
-		return(1);
+		return defb(-1);
 	}
 /*
 if((a>=0x40)&&(a<0x50)&&(ed40[a-0x40][0]!='0'))FP(fx,"%s",ed40[a-0x40]);
@@ -230,54 +251,63 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 	else if((a>=0xa0)&&(a<0xb0)&&(eda0[a-0xa0][0]!='0')) i=4;
 	else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0')) i=5;
 	else {
-		FP(fx,"defb 0edh");
+		defb(1);
 		FP(fx,";next byte illegal after ed");
 		return(1);
 	}
-	r=ckrange(2,bstr);
-	if(r==-1) {
-		FP(fx,"defb");
-		for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-		return(i);
-	}
 	f_z80=1;
 	switch(i) {
-		case 0: FP(fx,"%s",ed40[a-0x40]);
+		case 0: strcpy(rstr,ed40[a-0x40]);
 			break;
-		case 1: FP(fx,"%s",ed50[a-0x50]);
+		case 1: strcpy(rstr,ed50[a-0x50]);
 			break;
-		case 2: FP(fx,"%s",ed60[a-0x60]);
+		case 2: strcpy(rstr,ed60[a-0x60]);
 			break;
-		case 3: FP(fx,"%s",ed70[a-0x70]);
+		case 3: if(a < 0x72) {
+				before_undoc(2);
+			}
+			strcpy(rstr,ed70[a-0x70]);
 			break;
-		case 4: FP(fx,"%s",eda0[a-0xa0]);
+		case 4: strcpy(rstr,eda0[a-0xa0]);
 			break;
-		case 5: FP(fx,"%s",edb0[a-0xb0]);
+		case 5: strcpy(rstr,edb0[a-0xb0]);
 			break;
 	}
 
 /*if arg not in prog range it is a constant*/
 	if(a == 0x43)
 		{
-		FP(fx,"%s),bc",bstr);
-		return(4);
+		strcpy(stro,"%s),bc");
 		}
 	else if(a == 0x53)
 		{
-		FP(fx,"%s),de",bstr);
-		return(4);
+		strcpy(stro,"%s),de");
 		}
 	else if(a == 0x73)
 		{
-		FP(fx,"%s),sp",bstr);
-		return(4);
+		strcpy(stro,"%s),sp");
 		}
 	else if((a == 0x4b) || (a == 0x5b) || (a == 0x7b))
 		{
-		FP(fx,"%s)",bstr);
-		return(4);
+		strcpy(stro,"%s)");
 		}
-	return(2);
+	else {
+		stro[0] = 0;
+	}
+
+	if(stro[0] == 0) {
+		FP(fx,rstr);
+		return(2);
+	} else {
+		r=ckrange(2,bstr);
+		if(r==-1) {
+			return defb(-1);
+		} else {
+			FP(fx,rstr);
+			FP(fx,stro,bstr);
+		}
+		return(4);
+	}
 } else if((t[0] == 0xdd) || (t[0] == 0xfd)) {
 	if(t[0]==0xfd) {
 		strcpy(stri,"iy");
@@ -285,11 +315,16 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 		strcpy(stri,"ix");
 	}
 	if(t[1]==-1) {
-		FP(fx,"defb 0%02xh", t[0]);
-		return(1);
+		return defb(-1);
 	}
 	r=ckrange(2,bstr);
 	f_z80=1;
+
+	if(ddfd[t[1]][0]!='0') {
+		before_undoc(2);
+		FP(fx,ddfd[t[1]],stri);
+		return(2);
+	}
 	switch(t[1])
 		{
 		case 0x09:
@@ -302,18 +337,14 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 			break;
 		case 0x21:
 			if(r==-1) {
-				FP(fx,"defb");
-				for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-				return(i);
+				return defb(-1);
 			}
 			FP(fx,"ld %s,%s",stri,bstr);
 			return(4);
 			break;
 		case 0x22:
 			if(r==-1) {
-				FP(fx,"defb");
-				for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-				return(i);
+				return defb(-1);
 			}
 			FP(fx,"ld (%s),%s",bstr,stri);
 			return(4);
@@ -321,6 +352,13 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 		case 0x23:
 			FP(fx,"inc %s",stri);
 			return(2);
+		case 0x26:
+			if(t[2]==-1) {
+				return defb(-1);
+			}
+			before_undoc(3);
+			FP(fx,"ld %sh,0%02xh",stri,t[2]);
+			return(3);
 			break;
 		case 0x29:
 			FP(fx,"add %s,%s",stri,stri);
@@ -328,9 +366,7 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 			break;
 		case 0x2a:
 			if(r==-1) {
-				FP(fx,"defb");
-				for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-				return(i);
+				return defb(-1);
 			}
 			FP(fx,"ld %s,(%s)",stri,bstr);
 			return(4);
@@ -339,10 +375,38 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 			FP(fx,"dec %s",stri);
 			return(2);
 			break;
+		case 0x2e:
+			if(t[2]==-1) {
+				return defb(-1);
+			}
+			before_undoc(3);
+			FP(fx,"ld %sl,0%02xh",stri,t[2]);
+			return(3);
+			break;
 		case 0x39:
 			FP(fx,"add %s,sp",stri);
 			return(2);
 			break;
+		case 0x64:
+			before_undoc(2);
+			FP(fx,"ld %sh,%sh",stri,stri);
+			return(2);
+			break;
+		case 0x65:
+			before_undoc(2);
+			FP(fx,"ld %sh,%sl",stri,stri);
+			return(2);
+			break;
+		case 0x6c:
+			before_undoc(2);
+			FP(fx,"ld %sl,%sh",stri,stri);
+			return(2);
+			break;
+		case 0x6d:
+			before_undoc(2);
+			FP(fx,"ld %sl,%sl",stri,stri);
+			return(2);
+			break;
 		case 0xe1:
 			FP(fx,"pop %s",stri);
 			return(2);
@@ -367,9 +431,7 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 	/* now it must be an instruction of type (ix+dd) or (iy+dd) */
 
 	if((t[2]==-1)||(t[3]==-1)) {
-		FP(fx,"defb");
-		for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-		return(i);
+		return defb(-1);
 	}
 
 	if(t[2] < 128) {
@@ -381,107 +443,135 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 
 	if(t[1]==0xcb)
 		{
-		switch(t[3])
+		a = t[3] & 0x07;
+		if(a != 0x06) {
+			before_undoc(4);
+		}
+		switch(t[3] & 0xf8)
 			{
-			case 0x06:
+			case 0x00:
 				FP(fx,"rlc ");
 				break;
-			case 0x0e:
+			case 0x08:
 				FP(fx,"rrc ");
 				break;
-			case 0x16:
+			case 0x10:
 				FP(fx,"rl ");
 				break;
-			case 0x1e:
+			case 0x18:
 				FP(fx,"rr ");
 				break;
-			case 0x26:
+			case 0x20:
 				FP(fx,"sla ");
 				break;
-			case 0x2e:
+			case 0x28:
 				FP(fx,"sra ");
 				break;
-			case 0x3e:
+			case 0x30:
+				FP(fx,"sli ");
+				break;
+			case 0x38:
 				FP(fx,"srl ");
 				break;
-			case 0x46:
+			case 0x40:
 				FP(fx,"bit 0,");
 				break;
-			case 0x4e:
+			case 0x48:
 				FP(fx,"bit 1,");
 				break;
-			case 0x56:
+			case 0x50:
 				FP(fx,"bit 2,");
 				break;
-			case 0x5e:
+			case 0x58:
 				FP(fx,"bit 3,");
 				break;
-			case 0x66:
+			case 0x60:
 				FP(fx,"bit 4,");
 				break;
-			case 0x6e:
+			case 0x68:
 				FP(fx,"bit 5,");
 				break;
-			case 0x76:
+			case 0x70:
 				FP(fx,"bit 6,");
 				break;
-			case 0x7e:
+			case 0x78:
 				FP(fx,"bit 7,");
 				break;
-			case 0x86:
+			case 0x80:
 				FP(fx,"res 0,");
 				break;
-			case 0x8e:
+			case 0x88:
 				FP(fx,"res 1,");
 				break;
-			case 0x96:
+			case 0x90:
 				FP(fx,"res 2,");
 				break;
-			case 0x9e:
+			case 0x98:
 				FP(fx,"res 3,");
 				break;
-			case 0xa6:
+			case 0xa0:
 				FP(fx,"res 4,");
 				break;
-			case 0xae:
+			case 0xa8:
 				FP(fx,"res 5,");
 				break;
-			case 0xb6:
+			case 0xb0:
 				FP(fx,"res 6,");
 				break;
-			case 0xbe:
+			case 0xb8:
 				FP(fx,"res 7,");
 				break;
-			case 0xc6:
+			case 0xc0:
 				FP(fx,"set 0,");
 				break;
-			case 0xce:
+			case 0xc8:
 				FP(fx,"set 1,");
 				break;
-			case 0xd6:
+			case 0xd0:
 				FP(fx,"set 2,");
 				break;
-			case 0xde:
+			case 0xd8:
 				FP(fx,"set 3,");
 				break;
-			case 0xe6:
+			case 0xe0:
 				FP(fx,"set 4,");
 				break;
-			case 0xee:
+			case 0xe8:
 				FP(fx,"set 5,");
 				break;
-			case 0xf6:
+			case 0xf0:
 				FP(fx,"set 6,");
 				break;
-			case 0xfe:
+			case 0xf8:
 				FP(fx,"set 7,");
 				break;
-			default:
-		FP(fx,"defb 0%02xh,0%02xh,0%02xh,0%02xh\t;illegal sequence"\
-			,t[0],t[1],t[2],t[3]);
-				return(4);
 			}/*end switch t[3]*/
 		FP(fx,"%s",stro);
+		switch(a) {
+			case 0x00:
+				FP(fx," & ld b,%s",stro);
+				break;
+			case 0x01:
+				FP(fx," & ld c,%s",stro);
+				break;
+			case 0x02:
+				FP(fx," & ld d,%s",stro);
+				break;
+			case 0x03:
+				FP(fx," & ld e,%s",stro);
+				break;
+			case 0x04:
+				FP(fx," & ld h,%s",stro);
+				break;
+			case 0x05:
+				FP(fx," & ld l,%s",stro);
+				break;
+			case 0x06:
+				break;
+			case 0x07:
+				FP(fx," & ld a,%s",stro);
+				break;
+		}
 		return(4);
 		}/*end dd cb or fd cb*/
 	else switch(t[1])
@@ -563,8 +653,8 @@ else if((a>=0xb0)&&(a<0xc0)&&(edb0[a-0xb0][0]!='0'))FP(fx,"%s",edb0[a-0xb0]);
 			FP(fx,"cp %s",stro);
 			break;
 		default:
-			FP(fx,"defb 0%02xh,0%02xh,0%02xh\t;illegal sequence"\
-			,t[0],t[1],t[2]);
+			defb(3);
+			FP(fx,"\t;illegal sequence");
 			return(3);
 		}/*end switch t[1]*/
 	return(3);
@@ -578,25 +668,21 @@ switch(comtab[t[0]].type)
 		break;
 	case 1:
 		if(t[1]==-1) {
-			FP(fx,"defb 0%02xh", t[0]);
-			return(1);
+			return defb(-1);
 		}
 		FP(fx,"%s",comtab[t[0]].com1);
 		FP(fx,"0%02xh",t[1]);
 		break;
 	case 2:
 		if(r==-1) {
-			FP(fx,"defb");
-			for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-			return(i);
+			return defb(-1);
 		}
 		FP(fx,"%s",comtab[t[0]].com1);
 		FP(fx,"%s",bstr);
 		break;
 	case 3:
 		if(t[1]==-1) {
-			FP(fx,"defb 0%02xh", t[0]);
-			return(1);
+			return defb(-1);
 		}
 		FP(fx,"%s",comtab[t[0]].com1);
 		if(comtab[t[0]].bytes == 2) {
@@ -613,17 +699,14 @@ switch(comtab[t[0]].type)
 		break;
 	case 11:
 		if(t[1]==-1) {
-			FP(fx,"defb 0%02xh", t[0]);
-			return(1);
+			return defb(-1);
 		}
 		FP(fx,"%s",comtab[t[0]].com1);
 		FP(fx,"0%02xh%s",t[1],comtab[t[0]].com2);
 		break;
 	case 12:
 		if(r==-1) {
-			FP(fx,"defb");
-			for(i=0;t[i]!=-1;i++) FP(fx,"%c0%02xh", i?',':' ', t[i]);
-			return(i);
+			return defb(-1);
 		}
 		FP(fx,"%s",comtab[t[0]].com1);
 		FP(fx,"%s%s",bstr,comtab[t[0]].com2);
