@@ -13,6 +13,7 @@ struct comm{
 extern int a_address;
 extern int a_labels;
 extern int a_source;
+extern int a_undoc;
 extern int a_org;
 extern int a_zilog;
 
@@ -38,3 +39,4 @@ extern const char *ed60[];
 extern const char *ed70[];
 extern const char *eda0[];
 extern const char *edb0[];
+extern const char *ddfd[];
